from pathlib import Path
import pyarrow as pa
import nlp
from fastai.text.all import *

class HF_BaseTransform():
  "The base class of HuggingFace/nlp transform. Inherit it to get the ability to do :func:`map` on not only a :class:`nlp.Dataset` but also  :class:`nlp.Dataset` s at once."

  def __init__(self, hf_dsets, remove_original=False, out_cols=None):
    """
    Args:
      hf_dsets (Dict[ :class:`nlp.Dataset` ] or :class:`nlp.Dataset` ): The Hugging Face dataset(s) to do `map`
      remove_original (bool): whther to remove all original columns after `map`
      out_cols (List[str]): output column names. If specified, it will assure they are not in the columns to be removed.
    """
    # check arguments
    if isinstance(hf_dsets, nlp.Dataset): hf_dsets = {'Single': hf_dsets}; self.single = True
    else: self.single = False
    assert isinstance(hf_dsets, dict)
    # save attributes
    self.hf_dsets = hf_dsets
    self.remove_original,self.out_cols = remove_original,out_cols

  @property
  def cache_dir(self): return Path(next(iter(self.hf_dsets.values())).cache_files[0]['filename']).parent

  @delegates(nlp.Dataset.map, but=["cache_file_name"])
  def map(self, split_kwargs=None, cache_dir=None, cache_name=None, **kwargs):
    """
    Args:
      split_kwargs (Dict[dict] or List[dict]): arguments of :func:`_map` and :func:`nlp.Dataset.map` for specific splits.
      cache_dir (str, default=`None`): if ``None``, it is the cache dir of the (first) dataset. 
      cache_name (str, default=`None`): format string includes one param "{split}" that will be converted to split name, as cache file name under `cache_dir` for each split. If ``None``, use automatically generated hashed name by hf/nlp.  
      kwargs: passed to :func:`_map` and :func:`nlp.Dataset.map`   
    """
    # check/process arguments
    if self.remove_original: 
      assert 'remove_columns' not in kwargs, "You have specified to remove all original columns."
    if split_kwargs is None:
      split_kwargs = { split:{} for split in self.hf_dsets }
    elif isinstance(split_kwargs, list):
      split_kwargs = { split:split_kwargs[i] for i, split in enumerate(self.hf_dsets) }
    elif isinstance(split_kwargs, dict):
      for split in split_kwargs.keys(): assert split in self.hf_dsets, f"{split} is not the split names {list(self.hf_dsets.keys())}."
      for split in self.hf_dsets:
        if split not in split_kwargs: split_kwargs[split] = {}
    cache_dir = Path(cache_dir) if cache_dir else self.cache_dir
    cache_dir.mkdir(exist_ok=True)
    # map
    new_dsets = {}
    for split, dset in self.hf_dsets.items():
      if self.remove_original: kwargs['remove_columns'] = dset.column_names
      if cache_name: 
        if self.single: kwargs['cache_file_name'] = str(cache_dir/cache_name)
        else: kwargs['cache_file_name'] = str(cache_dir/cache_name.format(split=split))
      kwargs.update(split_kwargs[split])
      if hasattr(kwargs, 'remove_columns'): self._check_outcols(kwargs['remove_columns'], split)
      new_dsets[split] = self._map(dset, split, **kwargs)
    # return
    if self.single: return new_dsets['Single']
    else: return new_dsets

  def _check_outcols(self, out_cols, rm_cols, split):
    if not self.out_cols: return
    for col in self.out_cols: assert col not in rm_cols, f"Output column name {col} is in the list of columns {rm_cols} will be removed after `map`." + f"The split is {split}" if split != 'Single' else ''

  @delegates(nlp.Dataset.map)
  def _map(self, dset, split, **kwargs):
    """ Child class can override this method to implement. """
    return dset.map(self, **kwargs)
  
  # The method you need to implement
  def __call__(self, example): 
    """ Child class should implement this method, which takes an example (dict) and return an example (dict), just like ``function`` in :func:`nlp.Dataset.map`"""
    raise NotImplementedError

@delegates()
class HF_Transform(HF_BaseTransform):
  """ Like normal :func:`nlp.Dataset.map`, but is able to process multiple :class:`nlp.Dataset` """
  def __init__(self, hf_dset, func, **kwargs):
    """
    Args:
      hf_dset (:class:`nlp.Dataset` or Dict[ :class:`nlp.Dataset` ]): The Hugging Face dataset(s) to map
      func (`Callable[dict]->dict`): like `func` in `nlp.Dataset.map`
    
    Example:
      >>> rte = nlp.load_dataset('glue', 'rte')
      >>> def custom_tokenize(example):
      ...   example['tok_ids'] = hf_tokenizer.encode(example['sentence1'], example['sentence2'])
      ...   return example
      >>> tokenized_rte = HF_Transform(rte, custom_tokenize).map()
      >>> tokenized_rte['validation'][0]
      {'sentence1': 'Dana Reeve, the widow of the actor Christopher Reeve, has died of lung cancer at age 44, according to the Christopher Reeve Foundation.',
       'sentence2': 'Christopher Reeve had an accident.',
       'label': 1,
       'idx': 0,
       'tok_ids': [101, 11271, 20726, 1010, 1996, 7794, 1997, 1996, 3364, 5696, 20726, 1010, 2038, 2351, 1997, 11192, 4456, 2012, 2287, 4008, 1010, 2429, 2000, 1996, 5696, 20726, 3192, 1012, 102, 5696, 20726, 2018, 2019, 4926, 1012, 102]}
    """
    super().__init__(hf_dset, **kwargs)
    self.func = func
  def __call__(self, example): return self.func(example)

@delegates(but=["out_cols"])
class HF_TokenizeTfm(HF_BaseTransform):
  
  def __init__(self, hf_dset, cols, hf_toker, **kwargs):
    """
    Args:
      hf_dset (:class:`nlp.Dataset` or Dict[ :class:`nlp.Dataset` ]): The Hugging Face dataset(s) to do tokenization
      cols: with one of the following signature:\n
        - `cols`(`Dict[str]`): tokenize the every column named [key] into a column named [value]
        - `cols`(`List[str]`): specify the name of columns to be tokenized, replace the original columns' data with tokenized one
      hf_toker (:class:`transformers.PreTrainedTokenizer`, optional): Hugging Face tokenizer

    Example:
      >>> cola = nlp.load_dataset('glue', 'cola') 
      {'train':nlp.Dataset, 'validation':nlp.Dataset, 'test':nlp.Dataset}
      >>> cola['train][0]
      {'sentence': "Our friends won't buy this analysis, let alone the next one we propose.",
        'label': 1, 'idx': 0,}
      >>> tokenized_cola = HF_TokenizeTfm(cola, {'sentence':'text_idxs'}, hf_tokenizer).map()
      >>> tokenized_cola['train'][0]
      {'sentence': "Our friends won't buy this analysis, let alone the next one we propose.",
        'label': 1, 'idx': 0,
        'text_idxs': [2256, 2814, 2180, 1005, 1056, 4965, 2023, 4106, 1010, 2292, 2894, 1996, 2279, 2028, 2057, 16599, 1012]}
    """
    if isinstance(cols, list): cols = {c:c for c in cols}
    assert isinstance(cols, dict)
    super().__init__(hf_dset, out_cols=list(cols.values()), **kwargs)
    self.cols, self.tokenizer = cols, hf_toker
    """
    If don't specify cache file name, it will be hashed binary of pickled function that
    passed to `map`, so if you pass the same function, it knows to use cache.
    But tokenizer can't be pickled, so use tokenizer config to make tfms use different 
    tokenizers unique.  
    """
    self.tokenizer_config = hf_toker.pretrained_init_configuration
  
  def __call__(self, example):
    for in_col, out_col in self.cols.items():
      example[out_col] = self.tokenizer.convert_tokens_to_ids(self.tokenizer.tokenize(example[in_col]))
    return example

  def __getstate__(self):
    "specify something you don't want pickle here, remember to use copy to not modfiy orginal instance"
    state = self.__dict__.copy() 
    state['tokenizer'] = None 
    return state

class CombineTransform(HF_BaseTransform):
  """
  Base Class for Transform that combine multiple original samples into a new sample. 
  """
  def __init__(self, hf_dset, inp_cols, out_cols, init_attrs, drop_last=False):
    """
    Args:
      hf_dset (:class:`nlp.Dataset` or Dict[ :class:`nlp.Dataset` ]): The Hugging Face dataset(s) to do the transformation
      inp_cols (`List[str]`): names of input columns that used to produce samples
      out_cols (`List[str]`): names of output columns to put combined samples.
      init_attrs` (`List[str]`): name of attributes of children class that need to be their initial status when starts to aggregate dataset. i.e. Those defined in `__init__` and the value will changed during `accumulate`
      drop_last` (`Optional[bool]`, default: `False`): whether to drop the last accumulated sample.
    """
    super().__init__(hf_dset)
    self.inp_cols, self.out_cols =  inp_cols, out_cols
    for col in out_cols: assert col not in self.inp_cols, f"New column name can't be the same with any original column name. '{col}'"
    # batched map need dataset be in python format
    if isinstance(hf_dset, dict):
      for dset in hf_dset.values(): dset.set_format(type=None, columns=inp_cols) 
    else: hf_dset.set_format(type=None, columns=inp_cols)
    # dealing with last sample
    self.last_idx = len(hf_dset) - 1
    self.drop_last = drop_last
    # for reset
    self.init_attrs = init_attrs
    self.original_vals = [deepcopy(getattr(self, attr)) for attr in init_attrs]  

  def __call__(self, b, indices):
    # `nlp.Dataset.map` first test with several samples which affects our attrs, so we need to reinitialize.
    if 0 in indices: # reset
      for attr,val in zip(self.init_attrs, self.original_vals): setattr(self, attr, deepcopy(val))

    self.new_b = { c:[] for c in self.out_cols }
    for z in zip(*b.values()):
      self.accumulate(*z)
    
    # whehther put last example when it is last batch of `map`
    if not self.drop_last and self.last_idx in indices: 
      try: self.commit_example(self.create_example())
      except: pass # assume it is because there is nothing to create a example

    return self.new_b

  def commit_example(self, example):
    if example is None: return
    for col,val in example.items():
      self.new_b[col].append(val) 

  def accumulate(self, *args):
    """
    Child Class should implement this method.
    Given a example, do `self.commit_example(self.create_example()) when a new combined sample is ready.`
    Args:
      args : values of :data:`inp_cols` ( passed to :func:`__init__` ) of an example
    """
    raise NotImplementedError
  
  def create_example(self): 
    """
    Child Class should implement this method.
    Use attributes stored in the child class instance to create a combined sample (dict) when it is ready
    """
    raise NotImplementedError

  def _map(self, hf_dset, split, batch_size=1000, **kwargs):
    assert 'remove_columns' not in kwargs, "Aggregation type transform will only leave output columns for output dataset."
    return hf_dset.map(function=self, batched=True, batch_size=batch_size, with_indices=True, remove_columns=self.inp_cols, **kwargs)


@delegates(CombineTransform, but=["inp_cols", "out_cols", "init_attrs"])
class LMTransform(CombineTransform):
  """ Transform any dataset has tokenized text into dataset (autotgressive) language model. """
  def __init__(self, tokenized_hf_dset, max_len, text_col, x_text_col='x_text', y_text_col='y_text', **kwargs):
    """
    Args:
      tokenized_hf_dset (:class:`nlp.Dataset` or Dict[ :class:`nlp.Dataset` ]): tokenized Hugging Face dataset(s) to do LM transform 
      max_len (int): the length of a sentence
      text_col (str): the name of column that contains tokenized text (ids) of tokenized_hf_dset
      x_text_col (str): the name of the output column
      y_text_col (str): the name fo the output column
      kwargs: passed to :class:CombineTransform

    Example:
      >>> lm_dataset = LMTransform(tokenized_cola['validation'], max_len=20, text_col='text_idxs').map()
      >>> lm_dataset[0]
      {'x_text': [ 1996, 11279,  8469,  1996,  9478,  3154,  1997,  1996,  5749,  1012,
          1996, 15871,  2081,  1996,  8164,  7683,  2058,  1996,  4139,  3240],
       'y_text': [11279,  8469,  1996,  9478,  3154,  1997,  1996,  5749,  1012,  1996,
         15871,  2081,  1996,  8164,  7683,  2058,  1996,  4139,  3240,  1012]}
    """
    if isinstance(text_col, str): text_col = {text_col:['x_text','y_text']}
    assert isinstance(text_col, dict)
    self.text_col, (self.x_text_col, self.y_text_col) = next(iter(text_col.items()))
    self._max_len = max_len + 1
    self.residual_len, self.new_text = self._max_len, []
    super().__init__(tokenized_hf_dset, inp_cols=[self.text_col], out_cols=[x_text_col, y_text_col], init_attrs=['residual_len', 'new_text'], **kwargs)
    

  def accumulate(self, text): # *inp_cols
    """ Implement the abstract method"""
    usable_len = len(text)
    cursor = 0
    while usable_len != 0:
      use_len = min(usable_len, self.residual_len)
      self.new_text += text[cursor:cursor+use_len]
      self.residual_len -= use_len
      usable_len -= use_len
      cursor += use_len
      if self.residual_len == 0:
        self.commit_example(self.create_example())

  def create_example(self):
    """ Implement the abstract method"""
    # when read all data, the accumulated new_text might be less than two characters.
    if len(self.new_text) >= 2: 
      example = {self.x_text_col:self.new_text[:-1], self.y_text_col:self.new_text[1:]}
    else:
      example = None # mark "don't commit this"
    # reset accumulators
    self.new_text = []
    self.residual_len = self._max_len

    return example

@delegates(CombineTransform, but=["inp_cols", "out_cols", "init_attrs"])
class ELECTRADataTransform(CombineTransform):
  "Process any text corpus for ELECTRA's use"
  def __init__(self, hf_dset, is_docs, text_col, max_length, hf_toker, delimiter='\n', **kwargs):
    """
    Args:
      hf_dset (:class:`nlp.Dataset` or Dict[:class:`nlp.Dataset`]): **untokenized** Hugging Face dataset(s) to do the transform
      is_docs (bool): Whether each sample of this dataset is a doc
      text_col (str): the name of column of the dataset contains text 
      max_length (str): max length of each sentence
      hf_toker (:class:`transformers.PreTrainedTokenizer`): Hugging Face tokenizer
      delimiter (str): what is the delimiter to segment sentences in the input text
      kwargs: passed to :class:`CombineTransform`
    """
    self.is_docs = is_docs
    self.in_col = text_col
    self._current_sentences = []
    self._current_length = 0
    self._max_length = max_length
    self._target_length = max_length
    self.cls_idx, self.sep_idx = hf_toker.cls_token_id, hf_toker.sep_token_id
    self.hf_toker = hf_toker
    self.delimiter = delimiter
    super().__init__(hf_dset, inp_cols=[self.in_col], out_cols=['input_ids','sentA_lenth'], 
                    init_attrs=['_current_sentences', '_current_length', '_target_length'], **kwargs)

  """
  This two main functions adapts official source code creates pretraining dataset, to CombineTransform
  """
  def accumulate(self, text):
    sentences = text.split(self.delimiter)
    for sentence in sentences:
      if not sentence: continue # skip empty
      tokids = self.hf_toker.convert_tokens_to_ids(self.hf_toker.tokenize(sentence))
      self.add_line(tokids)
    # end of doc
    if self.is_docs and self._current_length > 0:
      self.commit_example(self.create_example())
  
  def create_example(self):
    input_ids, sentA_lenth = self._create_example() # this line reset _current_sentences and _current_length in the end
    return {'input_ids': input_ids, 'sentA_lenth':sentA_lenth}
  # ...................................................

  def add_line(self, tokids):
    """Adds a line of text to the current example being built."""
    self._current_sentences.append(tokids)
    self._current_length += len(tokids)
    if self._current_length >= self._target_length:
      self.commit_example(self.create_example())

  def _create_example(self):
    """Creates a pre-training example from the current list of sentences."""
    # small chance to only have one segment as in classification tasks
    if random.random() < 0.1:
      first_segment_target_length = 100000
    else:
      # -3 due to not yet having [CLS]/[SEP] tokens in the input text
      first_segment_target_length = (self._target_length - 3) // 2

    first_segment = []
    second_segment = []
    for sentence in self._current_sentences:
      # the sentence goes to the first segment if (1) the first segment is
      # empty, (2) the sentence doesn't put the first segment over length or
      # (3) 50% of the time when it does put the first segment over length
      if (len(first_segment) == 0 or
          len(first_segment) + len(sentence) < first_segment_target_length or
          (len(second_segment) == 0 and
           len(first_segment) < first_segment_target_length and
           random.random() < 0.5)):
        first_segment += sentence
      else:
        second_segment += sentence

    # trim to max_length while accounting for not-yet-added [CLS]/[SEP] tokens
    first_segment = first_segment[:self._max_length - 2]
    second_segment = second_segment[:max(0, self._max_length -
                                         len(first_segment) - 3)]

    # prepare to start building the next example
    self._current_sentences = []
    self._current_length = 0
    # small chance for random-length instead of max_length-length example
    if random.random() < 0.05:
      self._target_length = random.randint(5, self._max_length)
    else:
      self._target_length = self._max_length

    return self._make_example(first_segment, second_segment)

  def _make_example(self, first_segment, second_segment):
    """Converts two "segments" of text into a tf.train.Example."""
    input_ids = [self.cls_idx] + first_segment + [self.sep_idx]
    sentA_lenth = len(input_ids)
    if second_segment:
      input_ids += second_segment + [self.sep_idx]
    return input_ids, sentA_lenth

  def __getstate__(self):
    "specify something you don't want pickle here, remember to use copy to not modfiy orginal instance"
    state = self.__dict__.copy() 
    state['hf_toker'] = None 
    return state


from functools import partial
from pathlib import Path
import json
from tqdm import tqdm
from torch.nn.utils.rnn import pad_sequence
import nlp
from fastai.text.all import *


@delegates()
class MySortedDL(TfmdDL):
    "A :class:`DataLoader` that do smart batching and dynamic padding. Different from :class:`SortedDL`, it automatically pad every attribute of samples, is able to filter samples, and can be cached to sort/filter only at first time."

    def __init__(self, dataset, srtkey_fc=None, filter_fc=False, pad_idx=None, cache_file=None, **kwargs):
        """
        Args:
            dataset (HF_Dataset): Actually any object implements ``__len__`` and ``__getitem__`` that return a tuple as a sample.
            srtkey_fc (``*args->int``, optional): Get key for decending sorting from a sample .\n
              - If ``None``, sort by length of first element of a sample.
              - If ``False``, not sort. 
            filter_fc (``*args->bool``, optional): Return ``True`` to keep the sample.
            pad_idx (``int``, optional): pad each attribute of samples to the max length of its max length within the batch.\n 
              - If ``List[int]``, specify pad_idx for each attribute of a sample. e.g. a sample is a tuple (masked_inputs, labels), `pad_idx=[0 ,-100]` pad masked_inputs with 0, labels with -100.
              - If ``False``, do no padding. 
              - If ``None``, try ``dataset.pad_idx``, do no padding if no such attribute.
            cache_file (``str``, optional): Path of a json file to cache info for sorting and filtering.
            kwargs: key arguments for `TfmDl` or `DataLoader`

        Example:
            >>> samples = [ (torch.tensor([1]), torch.tensor([7,8]), torch.tensor(1)),,
            ...             (torch.tensor([2,3]), torch.tensor([9,10,11]), torch.tensor(2)),
            ...             (torch.tensor([4,5,6]), torch.tensor([11,12,13,14]), torch.tensor(3)), ]
            ... dl = MySortedDL(samples,
            ...                 srtkey_fc=lambda *args: len(args[0]),
            ...                 filter_fc=lambda x1,y1: y1<3,
            ...                 pad_idx=-1,
            ...                 cache_file='/tmp/cache.json', # calls after this will load cache
            ...                 bs=999, # other parameters go to `TfmDL` and `DataLoader`
            ...                 )
            ... dl.one_batch()
            (tensor([[ 2,  3],
                     [ 1, -1]]),
             tensor([[ 9, 10, 11],
                    [ 7,  8, -1]]),
             tensor([2, 1]))
        """
        # Defaults
        if srtkey_fc is not False: srtkey_fc = lambda *x: len(x[0])
        if pad_idx is None: pad_idx = getattr(dataset, 'pad_idx', False)
        if isinstance(pad_idx, int): pad_idxs = [pad_idx] * len(dataset[0])
        elif isinstance(pad_idx, (list, tuple)): pad_idxs = pad_idx
        cache_file = Path(cache_file) if cache_file else None
        idmap = list(range(len(dataset)))

        # Save attributes
        super().__init__(dataset, **kwargs)
        store_attr(self, 'pad_idxs,srtkey_fc,filter_fc,cache_file,idmap')

        # Prepare records for sorting / filtered samples
        if srtkey_fc or filter_fc:
          if cache_file and cache_file.exists():
            # load cache and check
            with cache_file.open(mode='r') as f: cache = json.load(f)
            idmap, srtkeys = cache['idmap'], cache['srtkeys']
            if srtkey_fc: 
              assert srtkeys, "srtkey_fc is passed, but it seems you didn't sort samples when creating cache."
              self.srtkeys = srtkeys
            if filter_fc:
              assert idmap, "filter_fc is passed, but it seems you didn't filter samples when creating cache."
              self.idmap = idmap
          else:
            # overwrite idmap if filter, get sorting keys if sort
            idmap = []; srtkeys = []
            for i in tqdm(range_of(dataset), leave=False):
                sample = self.do_item(i)
                if filter_fc and not filter_fc(*sample): continue
                if filter_fc: idmap.append(i)
                if srtkey_fc: srtkeys.append(srtkey_fc(*sample))
            if filter_fc: self.idmap = idmap
            if srtkey_fc: self.srtkeys = srtkeys
            # save to cache
            if cache_file:
              try: 
                with cache_file.open(mode='w+') as f: json.dump({'idmap':idmap,'srtkeys':srtkeys}, f)
              except: os.remove(str(cache_file))
          # an info for sorting
          if srtkey_fc: self.idx_max = np.argmax(self.srtkeys)
          # update number of samples
          if filter_fc: self.n = self.n = len(self.idmap)

    def create_item(self, i): return self.dataset[self.idmap[i]]

    def create_batch(self, samples):
        if self.pad_idx is False: return super().create_batch(samples)
        return tuple( pad_sequence(attr, batch_first=True, padding_value=self.pad_idxs[i]) if attr[0].shape and isinstance(self.pad_idxs[i], int) else torch.stack(attr) for i, attr in enumerate(zip(*samples)))

    def get_idxs(self):
        idxs = super().get_idxs()
        if self.shuffle: return idxs
        if self.srtkey_fc: return sorted(idxs, key=lambda i: self.srtkeys[i], reverse=True)
        return idxs

    def shuffle_fn(self,idxs):
        if not self.srtkey_fc: return super().shuffle_fn(idxs)
        idxs = np.random.permutation(self.n)
        idx_max = np.where(idxs==self.idx_max)[0][0]
        idxs[0],idxs[idx_max] = idxs[idx_max],idxs[0]
        sz = self.bs*50
        chunks = [idxs[i:i+sz] for i in range(0, len(idxs), sz)]
        chunks = [sorted(s, key=lambda i: self.srtkeys[i], reverse=True) for s in chunks]
        sort_idx = np.concatenate(chunks)

        sz = self.bs
        batches = [sort_idx[i:i+sz] for i in range(0, len(sort_idx), sz)]
        sort_idx = np.concatenate(np.random.permutation(batches[1:-1])) if len(batches) > 2 else np.array([],dtype=np.int)
        sort_idx = np.concatenate((batches[0], sort_idx) if len(batches)==1 else (batches[0], sort_idx, batches[-1]))
        return iter(sort_idx)

    @delegates(TfmdDL.new)
    def new(self, dataset=None, **kwargs):
        if 'get_idxs' in kwargs: # when Learner.get_preds, dataload has `get_idxs` will be cloned. So we need to prevent sorting again
          kwargs['cache_file'] = self.cache_file
        # We don't use filter_fc here cuz we can't don't validate certaion samples in dev/test set. 
        return super().new(dataset=dataset, pad_idx=self.pad_idx, srtkey_fc=self.srtkey_fc, filter_fc=False, **kwargs)

# =========================
#  Titled primitives
# =========================

class _Int(int, ShowPrint):
    def __new__(cls, *args, **kwargs):
        item = super().__new__(cls, *args)
        for n,v in kwargs.items(): setattr(item, n, v)
        return item

class _Float(float, ShowPrint):
    def __new__(cls, *args, **kwargs):
        item = super().__new__(cls, *args)
        for n,v in kwargs.items(): setattr(item, n, v)
        return item

class _Str(str, ShowPrint):
    def __new__(cls, *args, **kwargs):
        item = super().__new__(cls, *args)
        for n,v in kwargs.items(): setattr(item, n, v)
        return item

class _Tuple(fastuple, ShowPrint):
    def __new__(cls, *args, **kwargs):
        item = super().__new__(cls, *args)
        for n,v in kwargs.items(): setattr(item, n, v)
        return item 

class _L(L, ShowPrint):
    def __new__(cls, *args, **kwargs):
        item = super().__new__(cls, *args)
        for n,v in kwargs.items(): setattr(item, n, v)
        return item  

# only change "label" to "title"
def _show_title(o, ax=None, ctx=None, title=None, color='black', **kwargs):
    "Set title of `ax` to `o`, or print `o` if `ax` is `None`"
    ax = ifnone(ax,ctx)
    if ax is None: print(o)
    elif hasattr(ax, 'set_title'):
        t = ax.title.get_text()
        if len(t) > 0: o = t+'\n'+str(o)
        ax.set_title(o, color=color)
    elif isinstance(ax, pd.Series):
        while title in ax: title += '_'
        ax = ax.append(pd.Series({title: o}))
    return ax

class _ShowTitle:
    def show(self, ctx=None, **kwargs):
      kwargs['title'] = kwargs.pop('title', getattr(self, 'title', self.default_title))
      return _show_title(str(self), ctx=ctx, **kwargs)

# it seems that python prioritising prior inherited class when finding methods   

class _TitledInt(_ShowTitle, _Int): default_title = 'int'

class _TitledFloat(_ShowTitle, _Float): default_title = 'float'

# I created it, but it just print book likt int, haven't find a way to solve it
class _TitledBool(_ShowTitle, _Int): # python says bool can't be base class
    default_title = 'bool'

class _TitledStr(_ShowTitle, _Str):
    default_title = 'text'
    def truncate(self, n):
        "Truncate self to `n`"
        words = self.split(' ')[:n]
        return _TitledStr(' '.join(words), title=getattr(self, 'title', 'text'))

class _TitledTuple(_ShowTitle, _Tuple): default_title = 'list'

class _Category(_ShowTitle, _Str): default_title = 'label'

class _MultiCategory(_ShowTitle, _L):
    default_title = 'labels'
    def show(self, ctx=None, sep=';', color='black', **kwargs):
        kwargs['title'] = kwargs.pop('title', getattr(self, 'title', self.default_title))
        return __show_title(sep.join(self.map(str)), ctx=ctx, color=color, **kwargs)

""" Caution !!
These two function is inperfect.
But they cope with mutiple input columns problem (n_inp >1), which cause no df printing but just sequentail print
These will be a problem when you are doing non-text problem with n_inp > 1 (multiple input column),
which shouldn't be the case of huggingface/nlp user.
And I hope fastai come up with a good solution to show_batch multiple inputs problems for text/non-text.
"""
@typedispatch
def show_batch(x:tuple, y, samples, ctxs=None, max_n=9, **kwargs):
  if ctxs is None: ctxs = get_empty_df(min(len(samples), max_n))
  ctxs = show_batch[object](x, y, samples, max_n=max_n, ctxs=ctxs, **kwargs)
  display_df(pd.DataFrame(ctxs))
  return ctxs

@typedispatch
def show_results(x: tuple, y, samples, outs, ctxs=None, max_n=10, trunc_at=150, **kwargs):
  if ctxs is None: ctxs = get_empty_df(min(len(samples), max_n))
  ctxs = show_results[object](x, y, samples, outs, ctxs=ctxs, max_n=max_n, **kwargs)
  display_df(pd.DataFrame(ctxs))
  return ctxs

class HF_Dataset():
  """A wrapper for :class:`nlp.Dataset`.  It will behavior like original :class:`nlp.Dataset`, 
  but also function as a :class:`fastai.data.core.datasets` that provides samples and decodes."""
  
  def __init__(self, hf_dset, cols=None, hf_toker=None, neat_show=False, n_inp=1):
    """
    Args:
      hf_dset (:class:`nlp.Dataset`): Prerocessed Hugging Face dataset to be wrapped.
      cols (dict, optional): columns of :class:`nlp.Dataset` to be used to construct samples, and (optionally) semantic tensor type for each of those columns to decode.\n
        - cols(``Dict[Fastai Semantic Tensor]``): encode/decode column(key) with semantic tensor type(value). If {value} is ``noop``, semantic tensor of the column is by default `TensorTuple`.
        - cols(``list[str]``): specify only columns and take default setting for semantic tensor type of them.\n
          - if length is 1, regard the 1st element as `TensorText`
          - if length is 2, regard the 1st element as `TensorText`, 2nd element as `TensorCategory`
          - Otherwise, regard all elements as `TensorTuple`
        - cols(None): pass :data:`hf_dset.column_names` (list[str]) as cols.
      hf_toker (:class:`transformers.PreTrainedTokenizer`, optional): Hugging Face tokenizer, used in decode and provide ``pad_idx`` for dynamic padding
      neat_show (bool, optional): Show the original sentence instead of tokens joined by space.
      n_inp (int, optional): take the first ``n_inp`` columns of ``cols`` as x, and the rest as y .

    Example:
      >>> tokenized_cola_train_set[0]
      {'sentence': "Our friends won't buy this analysis, let alone the next one we propose.",
       'label': 1,
       'idx': 0,
       'text_idxs': [ 2256,  2814,  2180,  1005,  1056,  4965,  2023,  4106,  1010,  2292, 2894,  1996,  2279,  2028,  2057, 16599,  1012]}
      >>> hf_dset = HF_Datset(tokenized_cola_train_set, cols=['text_idxs', 'label'], hf_toker=tokenizer_electra_small_fast)
      >>> len(hf_dset), hf_dset[0]
      8551, (TensorText([ 2256,  2814,  2180,  1005,  1056,  4965,  2023,  4106,  1010,  2292, 2894,  1996,  2279,  2028,  2057, 16599,  1012]), TensorCategory(1))
      >>> hf_dset.decode(hf_dset[0])
      ("our friends won ' t buy this analysis , let alone the next one we propose .", '1')
      # The wrapped dataset "is" also the original huggingface dataset
      >>> hf_dset.column_names == tokenized_cola_train_set.column_names
      True
      # Manually specify `cols` with dict, here it is equivalent to the above. And addtionally, neatly decode samples.
      >>> neat_hf_dset = HF_Datset(tokenized_cola_train_set, {'text_idxs':TensorText, 'label':TensorCategory}, hf_toker=tokenizer_electra_small_fast, neat_show=True)
      >>> neat_hf_dset.decode(neat_hf_dset[0])
      ("our friends won't buy this analysis, let alone the next one we propose.", '1')
      # Note: Original set will be set to Pytorch format with columns specified in `cols`
      >>> tokenized_cola_train_set[0]
      {'label': tensor(1),
       'text_idxs': tensor([ 2256,  2814,  2180,  1005,  1056,  4965,  2023,  4106,  1010,  2292, 2894,  1996,  2279,  2028,  2057, 16599,  1012])}
    """
    
    # some default setting for tensor type used in decoding
    if cols is None: cols = hf_dset.column_names
    if isinstance(cols, list): 
      if n_inp==1: 
        if len(cols)==1: cols = {cols[0]: TensorText}
        elif len(cols)==2: cols = {cols[0]: TensorText, cols[1]: TensorCategory}
      else: cols = { c: noop for c in cols }
    assert isinstance(cols, dict)
    
    # make dataset output pytorch tensor
    if hf_dset.format['type'] != 'torch': 
      hf_dset.set_format( type='torch', columns=list(cols.keys()) )

    # store attributes
    self.pad_idx = hf_toker.pad_token_id
    store_attr(self, "hf_dset,cols,n_inp,hf_toker,neat_show")

  def __getitem__(self, idx):
    sample = self.hf_dset[idx]
    return tuple( tensor_cls(sample[col]) for col, tensor_cls in self.cols.items() )

  def __len__(self): return len(self.hf_dset)

  @property
  def col_names(self): return list(self.cols.keys())

  def decode(self, o, full=True): # `full` is for micmic `Dataset.decode` 
    if len(self.col_names) != len(o): return tuple( self._decode(o_) for o_ in o )
    return tuple( self._decode(o_, self.col_names[i]) for i, o_ in enumerate(o) )

  def _decode_title(self, d, title_cls, title): 
    if title: return title_cls(d, title=title)
    else: return title_cls(d)

  @typedispatch
  def _decode(self, t:torch.Tensor, title):
    if t.shape: title_cls = _TitledTuple
    elif isinstance(t.item(),bool): title_cls = _TitledBool # bool is also int, so check whether is bool first
    elif isinstance(t.item(),float): title_cls = _TitledFloat
    elif isinstance(t.item(),int): title_cls = _TitledInt
    return self._decode_title(t.tolist(), title_cls , title)

  @typedispatch
  def _decode(self, t:TensorText, title): 
    assert self.hf_toker, "You should give a huggingface tokenizer if you want to show batch."
    if self.neat_show: text = self.hf_toker.decode([idx for idx in t if idx != self.hf_toker.pad_token_id])
    else: text = ' '.join(self.hf_toker.convert_ids_to_tokens(t))
    return self._decode_title(text, _TitledStr, title)

  @typedispatch
  def _decode(self, t:LMTensorText, title): return self._decode[TensorText](self, t, title)

  @typedispatch
  def _decode(self, t:TensorCategory, title): return self._decode_title(t.item(), _Category, title)

  @typedispatch
  def _decode(self, t:TensorMultiCategory, title): return self._decode_title(t.tolist(), _MultiCategory, title)

  def __getattr__(self, name):
    "If not defined, let the nlp.Dataset in it act for us."
    if name in HF_Dataset.__dict__: return HF_Dataset.__dict__[name]
    elif name in self.__dict__: return self.__dict__[name]
    elif hasattr(self.hf_dset, name): return getattr(self.hf_dset, name)
    raise AttributeError(f"Both 'HF_Dataset' object and 'nlp.Dataset' object have no '{name}' attribute ")
  
class HF_Datasets(FilteredBase):
  """Function as :class:`fastai.data.core.Datasets` to create :class:`fastai.data.core.Dataloaders` from a group of :class:`nlp.Dataset`s"""

  _dl_type,_dbunch_type = MySortedDL,DataLoaders
  
  @delegates(HF_Dataset.__init__)
  def __init__(self, hf_dsets: dict, test_with_y=False, **kwargs):
    """
    Args:
      hf_dsets (`Dict[nlp.Dataset]`): Prerocessed Hugging Face Datasets, {key} is split name, {value} is :class:`nlp.Dataset`, order will become the order in :class:`fastai.data.core.Dataloaders`.
      test_with_y (bool, optional): Whether the test set come with y (answers) but not with fake y (e.g. all -1 label). 
        If ``False``, tell only test set to construct samples from first ``n_inp`` columns (do not output fake y). 
        And all datasets passed in ``hf_dsets`` with its name starts with "test" will be regarded as test set. 
      kwargs: Passed to :class:`HF_Dataset`. Be sure to pass arguments that :class:`HF_Dataset` needs !!
    """
    cols, n_inp = kwargs.pop('cols', None), kwargs.get('n_inp', 1)
    self.hf_dsets = {};
    for split, dset in hf_dsets.items():
      if cols is None: cols = dset.column_names
      if split.startswith('test') and not test_with_y: 
        if isinstance(cols, list): _cols = cols[:n_inp]
        else: _cols = { k:v for _, (k,v) in zip(range(n_inp),cols.items()) }
      else: _cols = cols
      self.hf_dsets[split] = HF_Dataset(dset, cols=_cols, **kwargs)

  def subset(self, i): return list(self.hf_dsets.values())[i]
  def __getitem__(self, split): return self.hf_dsets[split]
  @property
  def n_subsets(self): return len(self.hf_dsets)
  @property
  def cache_dir(self): return Path(next(iter(self.hf_dsets.values())).cache_files[0]['filename']).parent
  
  @delegates(FilteredBase.dataloaders)
  def dataloaders(self, device='cpu', cache_dir=None, cache_name=None, dl_kwargs=None, **kwargs):
    """
    Args:
      device (str): device where outputed batch will be on. Because a batch will be loaded to test when creating :class: `fastai.data.core.Dataloaders`, to prevent always leaving a batch of tensor in cuda:0, using default value cpu and then ``dls.to(other device)`` at the time you want is suggested.
      cache_dir (str, optional): directory to store caches of :class:`MySortedDL`. if ``None``, use cache directory of the first :class:`nlp.Dataset` in ``hf_dsets`` that passed to :method:`HF_Datasets.__init__`.
      cache_name (str, optional): format string that includes one param "{split}", which will be replaced with name of split as cache file name under `cache_dir` for each split. If ``None``, tell :class:MySortedDL don't do caching.
      dl_kwargs (list[dict], optional): ith item is addtional kwargs to be passed to initialization of ith dataloader for ith split
      kwargs: Passed to :func:`fastai.data.core.FilteredBase.dataloaders`
    
    Example:
      >>> tokenized_cola
      {'train': nlp.Dataset, 'validation': nlp.Dataset, 'test': nlp.Dataset}
      >>> tokenized_cola['test'][0]
      {'sentence': 'Bill whistled past the house.',
       'label': -1, # Fake label. True labels are not open to the public.
       'idx': 0,
       'text_idxs': [3021, 26265, 2627, 1996, 2160, 1012]}
      >>> dls = HF_Datasets(tokenized_cola,
      ...                   cols=['text_idxs', 'label'], hf_toker=hf_tokenizer,  # args for HF_Dataset
      ...                   ).dataloaders(bs=32 , cache_name="dl_cached_for_{split}") # args for MySortedDL
      >>> dls.show_batch(max_n=2)
                                                                                                                         text_idxs           label
      ---------------------------------------------------------------------------------------------------------------------------------------------
      0  everybody who has ever, worked in any office which contained any typewriter which had ever been used to type any letters which had    1
         to be signed by any administrator who ever worked in any department like mine will know what i mean.
      ---------------------------------------------------------------------------------------------------------------------------------------------
      1  playing with matches is ; lots of fun, but doing, so and emptying gasoline from one can to another at the same time is a sport best   1
         reserved for arsons.
      # test set won't produce label becuase of `test_with_y=False`   
      >>> dls[-1].show_batch(max_n=2) 
                                                                                     text_idxs
      ------------------------------------------------------------------------------------------
      0  cultural commissioner megan smith said that the five ` ` soundscape'' pieces would ` `
         give a festive air to park square, they're fun and interesting''.
      ------------------------------------------------------------------------------------------
      1  wendy is eager to sail around the world and bruce is eager to climb kilimanjaro, but 
         neither of them can because money is too tight.
    """
    if dl_kwargs is None: dl_kwargs = [{} for _ in range(len(self.hf_dsets))]
    elif isinstance(dl_kwargs, dict):
      dl_kwargs = [ dl_kwargs[split] if split in dl_kwargs else {} for split in self.hf_dsets]
    # infer cache file names for each dataloader if needed
    dl_type = kwargs.pop('dl_type', self._dl_type)
    if dl_type==MySortedDL and cache_name:
      assert "{split}" in cache_name, "`cache_name` should be a string with '{split}' in it to be formatted."
      cache_dir = Path(cache_dir) if cache_dir else self.cache_dir
      cache_dir.mkdir(exist_ok=True)
      if not cache_name.endswith('.json'): cache_name += '.json'
      for i, split in enumerate(self.hf_dsets):
        filled_cache_name = dl_kwargs[i].pop('cache_name', cache_name.format(split=split))
        if 'cache_file' not in dl_kwargs[i]:
          dl_kwargs[i]['cache_file'] = cache_dir/filled_cache_name
    # change default to not drop last
    kwargs['drop_last'] = kwargs.pop('drop_last', False)
    # when corpus like glue/ax has only testset, set it to non-train setting
    if list(self.hf_dsets.keys())[0].startswith('test'):
      kwargs['shuffle_train'] = False
      kwargs['drop_last'] = False
    return super().dataloaders(dl_kwargs=dl_kwargs, device=device, **kwargs)

def hf_merge_datasets(*datasets_s):
  """
  Args:
    *datasets_s: multiple dicts that contains :class:`nlp.Dataset`, each dict must have the same keys (split names), all datasets should have some columns with the same name.

  Returns
    Dict[ str : :class:`HF_MergeDataset` ]

  Example:
    >>> rte, wnli = nlp.load_dataset('glue', 'rte'), nlp.load_dataset('glue', 'wnli') # Just for example, you may not concates rte and wnli datasets in real.
    # rte: {'train':Dataset(schema:{...,'sentence1':...,'sentence2':...}),'validation':...}, wnli: {'train':Dataset(schema:{...,'sentence1':...,'sentence2':...}),'validation':...
    >>> merge_dsets = hf_merge_datasets(rte, wnli)
    {'train': HF_MergedDataset, 'validation': HF_MergedDataset, 'test': HF_MergedDataset}
  """
  keys_s = [ list(dsets.keys()) for dsets in datasets_s ]
  for keys in keys_s: assert keys == keys_s[0]
  merged_dsets = {}
  for split in keys:
    merged_dsets[split] = HF_MergedDataset(*[ dsets[split] for dsets in datasets_s])
  return merged_dsets

class HF_MergedDataset():
  """Merge multiple :class:`nlp.Dataset` s to be a fake :class:`nlp.Dataset` be able to passed to :class:`HF_Dataset`
  
  Args:
    *datasets : multiple :class:`nlp.Dataset` s, that all of these have some columns with the same names.

  Returns:
    HF_MergedDataset: a :class:`nlp.Dataset` like object that concats passed datasets, with basic functions to be turned into :class:`HF_Dataset`.

  Example:
    >>> tokenized_wiki_train, tokenized_bookcorpus_train
    Dataset(schema: {...., 'input_ids': 'list<item: int64>', ...), Dataset(schema: {...., 'input_ids': 'list<item: int64>', ...)
    >>> merged_dset = HF_MergedDataset(tokenized_wiki_train, tokenized_bookcorpus_train)
    >>> dls = HF_Datasets({'train': merged_dset}, cols=['input_ids'], hf_toker=hf_tokenizer).dataloaders(bs=128)
  """
  def __init__(self, *datasets):
    self.dsets = datasets
    self.len = reduce(lambda a,d: a+len(d), self.dsets, 0)
  def __len__(self):
    return self.len
  def __getitem__(self, i):
    for dset in self.dsets:
      if i < len(dset): return dset[i]
      else: i -= len(dset)
    raise IndexError
  def set_format(self, type, columns):
    for dset in self.dsets: dset.set_format(type, columns)
  @property
  def format(self):
    form = self.dsets[0].format
    for dset in self.dsets:
      assert form == dset.format
    return form
  @property
  def cache_files(self):
    return concat(*[ds.cache_files for ds in self.dsets])


import setuptools
from hugdatafast.__init__ import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

REQUIRED_PKGS = [
    'fastai',
    'nlp',
]

setuptools.setup(
    name="hugdatafast",
    version=__version__,
    author="Richard Wang",
    author_email="richardyy1188@gmail.com",
    description="The elegant bridge between hugginface data and fastai",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/richarddwang/hugdatafast",
    license='Apache 2.0',
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    python_requires='>=3.6',
    install_requires=REQUIRED_PKGS,
    keywords='nlp machine learning datasets metrics fastai huggingface',
)